----------------------------------------------------------------------
-- Galois package body (Galois.adb)
-- 
-- Defines funtions for point multiplication 
-- of Chapter 10 examples
---------------------------------------------------------------------
with Gnat.Io; use Gnat.Io;
package body Galois is
   function Divide_By_2(X, P: in Integer) return Integer is
      Y: Integer;
   begin
      if (X mod 2) = 0 then Y := X/2; else Y := (X+P)/2; end if;
      return Y;
   end Divide_By_2;
   function Divide_By_4(X, P: in Integer) return Integer is
      Y: Integer;
   begin
      if (P mod 4) = 1 then
         if (X mod 4) = 0 then Y := X/4;
         elsif (X mod 4) = 1 then Y := (X - P)/4;
         elsif (X mod 4) = 2 then Y := (X + 2*P)/4;
         else Y := (X + P)/4;
         end if;
      elsif (P mod 4) = 3 then
         if (X mod 4) = 0 then Y := X/4;
         elsif (X mod 4) = 1 then Y := (X + P)/4;
         elsif (X mod 4) = 2 then Y := (X + 2*P)/4;
         else Y := (X - P)/4;
         end if;
      end if;
      return Y;
   end Divide_By_4;   
   function Degree(A: Polynomial) return Natural is
      X: Natural;
   begin
      for I in 0 .. M loop if A(I) > 0 then X := I; end if; end loop;
      return X;
   end Degree; 
   function Subtract(A, B: Polynomial) return Polynomial is
      C: Polynomial;
   begin
      for I in 0 .. M loop C(I) := (A(I)-B(I)) mod P; end loop;
      return C;
   end Subtract;
   function Invert(Y: Natural_Mod_P) return Natural_Mod_P is
      A, B, C, D, Q, R, Next_D: Integer;
   begin
      A := P; B := Y; C := 0; D := 1;
      while B > 1 loop
         Q := A/B; R := A mod B; Next_D := C - D*Q;
         A := B; B := R; C := D; D := Next_D;
      end loop;
      return (D mod P);
   end Invert;
   function Product(A: Polynomial; B: Natural_Mod_P) return Polynomial is
      C: Polynomial;
   begin
      for I in 0 .. M loop C(I) := (A(I)*B) mod P; end loop;
      return C;
   end Product;
   function Shift_One(A: Polynomial) return Polynomial is
      B: Polynomial;
   begin
      for I in 0 .. M-1 loop B(I) := A(I+1); end loop;
      B(M) := 0;
      return B;
   end Shift_One;
   function Divide_By_X(A, F: Polynomial) return Polynomial is
   begin
      return Shift_One(Subtract(A, Product(F, ((A(0)*Invert(F(0))) mod P))));
   end Divide_By_X;
   function Add(A, B: Polynomial) return Polynomial is
      C: Polynomial;
   begin
      for I in 0 .. M loop C(I) := (A(I)+B(I)) mod P; end loop;
      return C;
   end Add;
   function Multiply_By_X(A, F: Polynomial) return Polynomial is
      C, D: Polynomial;
   begin
      for I in 1 .. M-1 loop C(I) := A(I-1);end loop;
      C(0) := 0; C(M) := 0;   
      for I in 0 .. M-1 loop D(I) := F(I);end loop;
      D(M) := 0;
      return Subtract(C, Product(D, A(M-1)));
   end Multiply_By_X;
   function Product_Mod_F(A, B, F: Polynomial) return Polynomial is
      C: Polynomial;
   begin
      for I in 0 .. M loop C(I) := 0; end loop;
      for I in 0 .. M-1 loop
         C := Add(Multiply_By_X(C, F), Product(A,B(M-1-I)));
      end loop;
      return C;
   end Product_Mod_F;
   function Frobenius(J, I: Natural)return Natural_Mod_P is 
      C: constant Natural := 2;
      T: Natural;   
      B, Z: Natural_Mod_P;
   begin
      T := P/M;
      B := (C**T) mod P;
      Z := 1;
      for K in 1 .. I*J loop Z := (Z*B)mod P; end loop;
      return Z;
   end Frobenius;
   function Quotient(Num, Den: Polynomial) return Polynomial is
      R, Q, A, B: Polynomial;
      U, V: Natural;
   begin
      R := Num;
      for I in 0 .. M loop Q(I) := 0; end loop;
      U := Degree(Num); V := Degree(Den);
      while U >= V loop
         Q(U-V) := (R(U)*Invert(Den(V))) mod P;
         A := Product(Den, Q(U-V));
         for I in U-V .. M loop B(I) := A(I-U+V); end loop;
         for I in 0 .. U-V-1 loop B(I) := 0; end loop;
         R := Subtract(R, B);
         U := Degree(R);
      end loop;
      return Q;
   end Quotient;
   function Remainder(Num, Den: Polynomial) return Polynomial is
      R, Q, A, B: Polynomial;
      U, V: Natural;
   begin
      R := Num;
      for I in 0 .. M loop Q(I) := 0; end loop;
      U := Degree(Num); V := Degree(Den);
      while U >= V loop
         Q(U-V) := (R(U)*Invert(Den(V))) mod P;
         A := Product(Den, Q(U-V));
         for I in U-V .. M loop B(I) := A(I-U+V); end loop;
         for I in 0 .. U-V-1 loop B(I) := 0; end loop;
         R := Subtract(R, B);
         U := Degree(R);
      end loop;
      return R;
   end Remainder;
   function Shift(A: Polynomial; T: Natural) return Polynomial is
      Shifted: Polynomial;
   begin
      for I in T .. M loop Shifted(I) := A(I-T); end loop;
      for I in 0 .. T-1 loop Shifted(I) := 0; end loop;                  
      return Shifted;
   end Shift;
   function Shift(A, F: Polynomial; T: Natural) return Polynomial is
      Shifted: Polynomial;
   begin
      Shifted := A;
      for J in 1 .. T loop
         Shifted := Multiply_By_X(Shifted, F);
      end loop;                  
      return Shifted;
   end Shift;
   procedure Swap(A, B: in out Polynomial) is
      Old_A: Polynomial;
   begin
      Old_A := A; A := B; B := Old_A;
   end Swap;
   procedure Swap(S, T: in out Natural) is 
      Old_S: Natural;
   begin
      Old_S := S; S := T; T := Old_S;
   end Swap;
   function Mp(X, Y: in Natural) return Natural is
      Product, Q, A, Z: Natural;
      X_Bin: Bit_Vector;
   begin
      Q := X; 
      for I in 0 .. Logp-1 loop X_Bin(I) := Q mod 2; Q := Q/2; end loop;     
      Product := 0;
      for I in 0 .. Logp-1 loop
         A := Product + X_Bin(I)*Y;
         Product := (A + (A mod 2)*P)/2;         
      end loop;
      if Product >= P then Z := Product - P; else Z := Product; end if;
      return Z;
   end Mp;
   function Divider_Mod_Binary_F(G, H, F: Polynomial) return Polynomial is
      A, B, C, D, Old_A, Old_C, Z: Polynomial;
      Dif, Min: Integer;
   begin
      A := H; B := F; C := G; 
      Dif := -1; Min := M-1;
      for I in 0 .. M loop D(I) := 0; end loop;
      while Min > 0 loop
         if A(0) = 0 then A := Shift_One(A); C := Divide_By_X(C, F); 
            if Dif <= 0 then Min := Min - 1; end if;
            Dif := Dif - 1;
         else 
            Old_A := A; Old_C := C;
            A := Shift_One(Subtract(A, B));
            C := Divide_By_X(Subtract(C, D),F);
            if Dif >= 0 then 
               if Dif = 0 then Min := Min-1; end if; 
               Dif := Dif - 1;
            else
               Dif := -Dif - 1;
               B := Old_A; D := Old_C;
            end if;
         end if;      
      end loop;
      if A(0) = 0 then Z := D; else Z := C; end if;
      return Z;
   end Divider_Mod_Binary_F;

   --procedure hex_to_bin(a: in hex_polynomial; b: out polynomial) is
   --begin
   --for i in 0 .. Mdiv4 loop
   --   b(M - (4*i + 3)) := a(i) mod 2;
   --   b(M - (4*i + 2)) := (a(i)/2) mod 2;
   --   b(M - (4*i + 1)) := (a(i)/4) mod 2;
   --   b(M - (4*i)) := (a(i)/8) mod 2;
   --end loop;
   --end hex_to_bin;  

   --procedure bin_to_hex(a: in polynomial; b: out hex_polynomial) is
   --begin
   --for i in 0 .. Mdiv4 loop
   --   b(i) := 8*a(M - (4*i)) + 4*a(M - (4*i + 1)) + 2*a(M - (4*i + 2)) + a(M - (4*i+ 3));
   --end loop; 
   --end bin_to_hex;

   function Hex_To_Bin(A: Hex_Polynomial) return Polynomial is
      B: Polynomial;
      --c: hex_polynomial;
   begin
      for I in 0 .. Mdiv4 loop
         --   b(M - (4*i + 3)) := a(i) mod 2;
         --   b(M - (4*i + 2)) := (a(i)/2) mod 2;
         --   b(M - (4*i + 1)) := ((a(i)/2)/2) mod 2;
         --   b(M - (4*i)) := (((a(i)/2)/2)/2) mod 2;
         B(4*I) := A(Mdiv4 - I) mod 2;
         B(4*I + 1) := (A(Mdiv4 - I)/2) mod 2;
         B(4*I + 2) := ((A(Mdiv4 - I)/2)/2) mod 2;
         B(4*I + 3) := (((A(Mdiv4 - I)/2)/2)/2) mod 2;
      end loop;
      return B;
   end Hex_To_Bin;  

   function Bin_To_Hex(A: Polynomial) return Hex_Polynomial is
      B: Hex_Polynomial;
   begin
      for I in 0 .. Mdiv4 loop
         B(I) := 8*A(M - (4*I)) + 4*A(M - (4*I + 1)) + 2*A(M - (4*I + 2)) + A(M - (4*I+ 3));
      end loop; 
      return B;
   end Bin_To_Hex;

   --function point_addition_x(xP, yP, xQ, yQ, f: Polynomial) return Polynomial is
   --xR, op1, op2, lambda, square_lambda: Polynomial;
   --begin
   --op1 := add(yP, yQ);
   --op2 := add(xP, xQ);
   --lambda := divider_mod_binary_f(op1, op2, f);
   --square_lambda := product_mod_f(lambda, lambda, f);
   --op1 := add(square_lambda, lambda);
   --op1 := add(op1, xP);
   --xR := add(op1, xQ);
   --xR(0) := 1 - xR(0);
   --return xR;
   --end point_addition_x;

   --function point_addition_y(xP, yP, xQ, yQ, f: Polynomial) return Polynomial is
   --xR, yR, op1, op2, lambda, square_lambda: Polynomial;
   --begin
   --op1 := add(yP, yQ);
   --op2 := add(xP, xQ);
   --lambda := divider_mod_binary_f(op1, op2, f);
   --square_lambda := product_mod_f(lambda, lambda, f);
   --op1 := add(square_lambda, lambda);
   --op1 := add(op1, xP);
   --xR := add(op1, xQ);
   --xR(0) := 1 - xR(0);
   --op1 := add(xP, xR);
   --op1 := product_mod_f(op1, lambda, f);
   --op1 := add(op1, xR);
   --yR := add(op1, yP);   
   --return yR;
   --end point_addition_y;

   procedure Point_Addition(Xp, Yp, Xq, Yq, F: in Polynomial; Xr, Yr: out Polynomial) is
      Op1, Op2, Lambda, Square_Lambda: Polynomial;
   begin
      Op1 := Add(Yp, Yq);
      Op2 := Add(Xp, Xq);
      Lambda := Divider_Mod_Binary_F(Op1, Op2, F);
      Square_Lambda := Product_Mod_F(Lambda, Lambda, F);
      Op1 := Add(Square_Lambda, Lambda);
      Op1 := Add(Op1, Xp);
      Xr := Add(Op1, Xq);
      Xr(0) := 1 - Xr(0);
      Op1 := Add(Xp, Xr);
      Op1 := Product_Mod_F(Op1, Lambda, F);
      Op1 := Add(Op1, Xr);
      Yr := Add(Op1, Yp);   
   end Point_Addition;

   function Point_Addition(Xa, Xb, Xp, F: Polynomial) return Polynomial is
      Xr, Square: Polynomial;
   begin
      Xr := Add(Xa, Xb);
      Xr := Divider_Mod_Binary_F(Xb, Xr, F);
      Square := Product_Mod_F(Xr, Xr, F);
      Xr := Add(Xr, Square);
      Xr := Add(Xp, Xr);
      return Xr;
   end Point_Addition;

   --function point_doubling_x(xP, yP, f: Polynomial) return Polynomial is
   --xR, op1, lambda, square_lambda: Polynomial;
   --begin
   --op1 := divider_mod_binary_f(yP, xP, f);
   --lambda := add(xP, op1);
   --square_lambda := product_mod_f(lambda, lambda, f);
   --xR := add(square_lambda, lambda);
   --xR(0) := 1 - xR(0);
   --return xR;
   --op1 := product_mod_f(lambda, xR, f);
   --op2 := product_mod_f(xP, xP, f);
   --op1 := add(op1, op2);
   --yR := add(op1, xR);   
   --end point_doubling_x;

   --function point_doubling_y(xP, yP, f: Polynomial) return Polynomial is
   --xR, yR, op1, op2, lambda, square_lambda: Polynomial;
   --begin
   --op1 := divider_mod_binary_f(yP, xP, f);
   --lambda := add(xP, op1);
   --square_lambda := product_mod_f(lambda, lambda, f);
   --xR := add(square_lambda, lambda);
   --xR(0) := 1 - xR(0);
   --op1 := product_mod_f(lambda, xR, f);
   --op2 := product_mod_f(xP, xP, f);
   --op1 := add(op1, op2);
   --yR := add(op1, xR);   
   --return yR;
   --end point_doubling_y;

   procedure Point_Doubling(Xp, Yp, F: in Polynomial; Xr, Yr: out Polynomial) is
      Op1, Op2, Lambda, Square_Lambda: Polynomial;
   begin
      Op1 := Divider_Mod_Binary_F(Yp, Xp, F);
      Lambda := Add(Xp, Op1);
      Square_Lambda := Product_Mod_F(Lambda, Lambda, F);
      Xr := Add(Square_Lambda, Lambda);
      Xr(0) := 1 - Xr(0);
      Op1 := Product_Mod_F(Lambda, Xr, F);
      Op2 := Product_Mod_F(Xp, Xp, F);
      Op1 := Add(Op1, Op2);
      Yr := Add(Op1, Xr);   
   end Point_Doubling;


   function Point_Doubling(Xa, F: Polynomial) return Polynomial is
      Xr, Inverse, Unit: Polynomial;
   begin
      Unit(0) := 1; for I in 1 .. M loop Unit(I) := 0; end loop;
      Xr := Product_Mod_F(Xa, Xa, F);
      Inverse := Divider_Mod_Binary_F(Unit, Xr, F);
      Xr := Add(Xr, Inverse);
      return Xr;
   end Point_Doubling;

   function Compute_Y(Xa, Xb, Xp, Yp, F: Polynomial) return Polynomial is
      Op1, Op2, Op3: Polynomial;
   begin
      Op1 := Add(Xa, Xp);
      Op2 := Add(Xb, Xp);
      Op3 := Product_Mod_F(Xp, Xp, F);
      Op2 := Product_Mod_F(Op1, Op2, F);
      Op2 := Add(Op2, Op3);
      Op2 := Add(Op2, Yp);
      Op2 := Product_Mod_F(Op1, Op2, F);
      Op2 := Divider_Mod_Binary_F(Op2, Xp, F);
      Op2 := Add(Op2, Yp);
      return Op2;
   end Compute_Y;

   procedure Point_Multiplication(K, Xp, Yp, F: in Polynomial; Xr, Yr: out Polynomial) is
      Int_K, Xq, Yq, Xxp, Yyp, New_Xq, New_Yq, New_Xxp, New_Yyp: Polynomial;
      Q_Infinity: Boolean;
   begin
      Int_K := K;
      --xQ := infinity; yQ := infinity; 
      Q_Infinity := True;
      Xxp := Xp; Yyp := Yp;
      for I in 0 .. M loop
         if Int_K(I) = 1 then
            --if (xxP = infinity) and (yyP = infinity) then null;
            --elsif (xQ = infinity) and (yQ = infinity) then xQ := xxP; yQ := yyP;
            if Q_Infinity then Xq := Xxp; Yq := Yyp; Q_Infinity := False;
               --put("Q_infinity"); New_Line;
               --elsif (xQ = xxP) and (yQ = add(xxP, yyP)) then xQ := infinity; yQ := infinity;
               --elsif (xQ = xxP) and (yQ = add(xxP, yyP)) then Q_infinity := true;
               --put("Q = -P"); New_Line;
               --elsif (xxP = xQ) and (yyP = yQ) then point_doubling(xxP, yyP, f, xQ, yQ); Q_infinity := false;
               --put("Q = P"); New_Line;
            else Point_Addition(Xxp, Yyp, Xq, Yq, F, New_Xq, New_Yq); Xq := New_Xq; Yq := New_Yq; 
               --Q_infinity := false;
            end if;
         end if;
         Point_Doubling(Xxp, Yyp, F, New_Xxp, New_Yyp); Xxp := New_Xxp; Yyp := New_Yyp;
      end loop;
      if Q_Infinity then Xr := Infinity; Yr := Infinity; else Xr := Xq; Yr := Yq; end if;
      --xR := xQ; yR := yQ;
   end Point_Multiplication;

   procedure Base_Tau_Conversion(K: in Polynomial; Z: out Signed_Bit_Vector; T: out Natural) is
      A, Old_A, B, Zero: Polynomial;
      --z: signed_bit_vector(0 .. 512);
      Carries: Carry_Vector;
      I: Natural;
   begin
      A := K;
      for J in 0 .. M loop B(J) := 0; end loop;
      I := 0;
      for J in 0 .. M loop Zero(J) := 0; end loop;
      while ((A /= Zero) or (B /= Zero)) loop 

         if I > 512 then Put("i greater than 512"); New_Line; exit; end if;

         if A(0) = 0 then Z(I) := 0; Carries(0) := 0;
         elsif A(1) = B(0) then Z(I) := 1; Carries(0) := 0;
         else Z(I) := -1; Carries(0) := 1;
         end if;
         Old_A := A;
         for J in 0 .. M-1 loop
            if Carries(J) + B(J) + Old_A(J+1) >= 2 then Carries(J+1) := 1; else Carries(J+1) := 0; end if; 
            A(J) := (Carries(J) + B(J) + Old_A(J+1)) mod 2;
         end loop;
         A(M) := (Carries(M) + B(M) + Old_A(M)) mod 2;
         Carries(0) := 1 - Carries(0);
         for J in 1 .. M loop Old_A(J) := 1 - Old_A(J); end loop;
         for J in 0 .. M-1 loop
            if Carries(J) + Old_A(J+1) = 2 then Carries(J+1) := 1; else Carries(J+1) := 0; end if; 
            B(J) := (Carries(J) + Old_A(J+1)) mod 2;
         end loop;
         B(M) := (Carries(M) + Old_A(M)) mod 2;
         I := I+1;
      end loop;

      --return z;
      T := I;

   end Base_Tau_Conversion;

   procedure Frobenius_Point_Multiplication(K, Xp, Yp, F: in Polynomial; Xr, Yr: out Polynomial)  is
      Xxp, Yyp, Xq, Yq, New_Xq, New_Yq: Polynomial;
      Z: Signed_Bit_Vector(0 .. 512);
      T: Natural;

   begin
      Base_Tau_Conversion(K, Z, T);
      Xq := Infinity; Yq := Infinity; 
      Xxp := Xp; Yyp := Yp;
      for I in 0 .. T-1 loop
         if Z(I) = 1 then
            --compute Q+P:
            if (Xxp = Infinity) and (Yyp = Infinity) then null;
            elsif (Xq = Infinity) and (Yq = Infinity) then Xq := Xxp; Yq := Yyp;
            elsif (Xq = Xxp) and (Yq = Add(Xxp, Yyp)) then Xq := Infinity; Yq := Infinity;
            elsif (Xxp = Xq) and (Yyp = Yq) then Point_Doubling(Xxp, Yyp, F, Xq, Yq);
            else Point_Addition(Xxp, Yyp, Xq, Yq, F, New_Xq, New_Yq); Xq := New_Xq; Yq := New_Yq;
            end if;
         elsif Z(I) = -1 then
            --compute Q-P
            if (Xxp = Infinity) and (Yyp = Infinity) then null; -- Q:= Q-P = Q
            elsif (Xq = Infinity) and (Yq = Infinity) then Xq := Xxp; Yq := Add(Xxp, Yyp); -- Q := Q-P = -P
            elsif (Xq = Xxp) and (Yq = Add(Xxp, Yyp)) then Point_Doubling(Xq, Yq, F, New_Xq, New_Yq); Xq := New_Xq; Yq := New_Yq; -- Q := Q-P = 2P
            elsif (Xxp = Xq) and (Yyp = Yq) then Xq := Infinity; Yq := Infinity; -- Q := Q-P = 0;
            else Point_Addition(Xxp, Add(Xxp, Yyp), Xq, Yq, F, New_Xq, New_Yq); Xq := New_Xq; Yq := New_Yq;
            end if;
         end if;
         --compute tau(P)
         Xxp := Product_Mod_F(Xxp, Xxp, F); Yyp := Product_Mod_F(Yyp, Yyp, F);    
      end loop;
      Xr := Xq; Yr := Yq;
   end Frobenius_Point_Multiplication;

   procedure Frobenius_Point_Multiplication_2(K, Xp, Yp, F: in Polynomial; Xr, Yr: out Polynomial)  is
      A, Old_A, B, Zero: Polynomial;
      Carries: Carry_Vector;
      Xxp, Yyp, Xq, Yq, New_Xq, New_Yq: Polynomial;
      Q_Infinity: Boolean;
   begin
      Q_Infinity := True;
      Xxp := Xp; Yyp := Yp;
      for J in 0 .. M loop Zero(J) := 0; end loop;
      A := K; B := Zero;
      while ((A /= Zero) or (B /= Zero)) loop 
         if A(0) = 0 then 
            Carries(0) := 0;
         elsif A(1) = B(0) then 
            Carries(0) := 0;
            if Q_Infinity then Xq := Xxp; Yq := Yyp; Q_Infinity := False;
            else Point_Addition(Xxp, Yyp, Xq, Yq, F, New_Xq, New_Yq); Xq := New_Xq; Yq := New_Yq;
            end if;
         else 
            Carries(0) := 1;
            if Q_Infinity then Xq := Xxp; Yq := Add(Xxp, Yyp); Q_Infinity := False; -- Q := Q-P = -P
            else Point_Addition(Xxp, Add(Xxp, Yyp), Xq, Yq, F, New_Xq, New_Yq); Xq := New_Xq; Yq := New_Yq;
            end if;
         end if;
         --compute tau(P)
         Xxp := Product_Mod_F(Xxp, Xxp, F); Yyp := Product_Mod_F(Yyp, Yyp, F);    
         --update a and b
         Old_A := A;
         for J in 0 .. M-1 loop
            if Carries(J) + B(J) + Old_A(J+1) >= 2 then Carries(J+1) := 1; else Carries(J+1) := 0; end if; 
            A(J) := (Carries(J) + B(J) + Old_A(J+1)) mod 2;
         end loop;
         A(M) := (Carries(M) + B(M) + Old_A(M)) mod 2;
         Carries(0) := 1 - Carries(0);
         for J in 1 .. M loop Old_A(J) := 1 - Old_A(J); end loop;
         for J in 0 .. M-1 loop
            if Carries(J) + Old_A(J+1) = 2 then Carries(J+1) := 1; else Carries(J+1) := 0; end if; 
            B(J) := (Carries(J) + Old_A(J+1)) mod 2;
         end loop;
         B(M) := (Carries(M) + Old_A(M)) mod 2;
      end loop;
      Xr := Xq; Yr := Yq;
   end Frobenius_Point_Multiplication_2;

   procedure Montgomery_Point_Multiplication (K, Xp, Yp, F: in Polynomial; Xr, Yr: out Polynomial)  is
      Xa, Xb: Polynomial;
      A_Infinity, B_Infinity: Boolean;
   begin
      Xb := Xp; A_Infinity := True; B_Infinity := False;
      for I in 1 .. M loop
         if K(M-I) = 0 then
            if A_Infinity then null; 
               --elsif B_infinity then 
               --xB := xA; B_infinity := false;
               --if xA = infinity then A_infinity := true; else xA := point_doubling(xA, f); end if;
            elsif Xb = Xa then 
               B_Infinity := True; 
               --if xA = infinity then A_infinity := true; else xA := point_doubling(xA, f); end if;
               Xa := Point_Doubling(Xa, F);
            else 
               Xb := Point_Addition(Xa, Xb, Xp, F); 
               --if xA = infinity then A_infinity := true; else xA := point_doubling(xA, f); end if;
               Xa := Point_Doubling(Xa, F); 
               --A_infinity := false;
            end if;
         else 
            --if B_infinity then null;
            --elsif A_infinity then 
            if A_Infinity then
               Xa := Xb; A_Infinity := False; 
               --if xB = infinity then B_infinity := true; else xB := point_doubling(xB, f); end if;
               Xb := Point_Doubling(Xb, F);
               --elsif xB = xA then 
               --A_infinity := true; 
               --if xB = infinity then B_infinity := true; else xB := point_doubling(xB, f); end if;
               --xB := point_doubling(xB, f);
            else
               Xa := Point_Addition(Xa, Xb, Xp, F); 
               --A_infinity := false;
               --if xB = infinity then B_infinity := true; else xB := point_doubling(xB, f); end if;             
               Xb := Point_Doubling(Xb, F);
            end if;
         end if;
      end loop;
      if A_Infinity then Xr := Infinity; Yr := Infinity;
      elsif B_Infinity then Xr := Xp; Yr := Add(Xp, Yp);
      else Xr := Xa; Yr := Compute_Y(Xa, Xb, Xp, Yp, F); end if;
   end Montgomery_Point_Multiplication;


end Galois;

